/*
 * Copyright (c) Members of the EGEE Collaboration. 2009-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#ifndef _CAPI_SD_PROTOS_H_
#define _CAPI_SD_PROTOS_H_

#include "sd_types.h"

#ifdef __cplusplus
extern "C" {
#endif

/* This function will initialise a DISCOVERER struct
 * with sensible defaults and the information provider URL if provided.
 * On error errorStr will, if suplied, be populated with an error string.
 * Do not use a DISCOVERER struct before calling this function on it.
 */
SD_RETURN_TYPE InitDiscoverer(DISCOVERER* disco,
                              const char* url,
                              char** errorStr);

/* This function will free/release the internal data
 * of a DISCOVERER struct.  The DISCOVERER struct
 * can not be used again until InitDiscoverer() has
 * been called on it.
 * On error errorStr will, if suplied, be populated with an error string.
 * This function DOES NOT free the DISCOVERER struct
 * if it has been dynamically allocated.  That MUST
 * be done seperately.
 */
SD_RETURN_TYPE ReleaseDiscoverer(DISCOVERER* disco, char** errorStr);

/* This function will call list_services on the supplied
 * DISCOVERER struct.
 * This is the '2' argument version taking a
 * service filter and a data filter.
 * On error errorStr will, if suplied, be populated with an error string.
 * The results will be returned in the supplied SERVICE_DESCRIPTION_ARRAY
 * Thiss array will need to be released with a call
 * to ReleaseServiceDescriptionArray() when the user has finished with it
 */
SD_RETURN_TYPE ListServices2(const DISCOVERER* disco,
                             const char* serviceFilter,
                             const char* dataFilter,
                             SERVICE_DESCRIPTION_ARRAY* arr,
                             char** errorStr);

/* This function will call list_services on the supplied
 * DISCOVERER struct.
 * This is the '3' argument version taking a
 * service filter and a data filter.
 * On error errorStr will, if suplied, be populated with an error string.
 * The results will be returned in the supplied SERVICE_DESCRIPTION_ARRAY
 * This array will need to be released with a call
 * to ReleaseServiceDescriptionArray() when the user has finished with it
 */
SD_RETURN_TYPE ListServices3(const DISCOVERER* disco,
                             const char* serviceFilter,
                             const char* authzlter,
                             const char* dataFilter,
                             SERVICE_DESCRIPTION_ARRAY* arr,
                             char** errorStr);

/* This function will free/release the internal data contained
 * within a SERVICE_DESCRIPTION_ARRAY struct.  All SERVICE_DESCRIPTION
 * structs that it contains will be freed/released so make copies
 * of any data you want to keep.
 * On error errorStr will, if suplied, be populated with an error string.
 */
SD_RETURN_TYPE ReleaseServiceDescriptionArray(SERVICE_DESCRIPTION_ARRAY* arr,
                                              char** errorStr);

/* This function will return a copy of the given SERVICE_DESCRIPTION struct.
 * This is the only acceptable method of copying such a structure.
 * On error errorStr will, if suplied, be populated with an error string.
 * The returned structure will be allocated on the heap and therefore
 * ReleaseServiceDescription() followed by free() must be called
 * on it when you have finished using it.
 */
SERVICE_DESCRIPTION* CopyServiceDescription(const SERVICE_DESCRIPTION* desc,
                                            char** errorStr);

/* This function will free/release the internal data contained
 * within a SERVICE_DESCRIPTION struct.  The SERVICE_DATA
 * struct that it contains will be freed/released so make copies
 * of any data you want to keep.
 * On error errorStr will, if suplied, be populated with an error string.
 */
SD_RETURN_TYPE ReleaseServiceDescription(SERVICE_DESCRIPTION* desc,
                                         char** errorStr);

/* This function returns an ATTRIBUTE_NAMES structure for the
 * requested SERVICE_DESCRIPTION
 * On error errorStr will, if suplied, be populated with an error string.
 * The results will be returned in the supplied ATTRIBUTE_NAMES
 * This structure will need to be released with a call
 * to ReleaseAttributesArray() when the user has finished with it
 */
SD_RETURN_TYPE ServiceDescriptionListAttributes(SERVICE_DESCRIPTION* desc,
                                                ATTRIBUTE_NAMES* arr,
                                                char** errorStr);

/* This function will free/release the internal data contained
 * within an ATTRIBUTE_NAMES struct.  All stings that it contains
 * will be freed so make copies of any data you want to keep.
 * On error errorStr will, if suplied, be populated with an error string.
 */
SD_RETURN_TYPE ReleaseAttributeNames(ATTRIBUTE_NAMES* arr,
                                     char** errorStr);

/* This function returns all the values of a given attribute
 * for a given SERVICE_DESCRIPTION
 * On error errorStr will, if suplied, be populated with an error string.
 * The results will be returned in the supplied ATTRIBUTE_VALUES
 * This structure will need to be released with a call
 * to ReleaseAttributeValues() when the user has finished with it
 */
SD_RETURN_TYPE ServiceDescriptionGetAttribute(SERVICE_DESCRIPTION* desc,
                                              const char* attribute,
                                              ATTRIBUTE_VALUES* values,
                                              char** errorStr);

/* This function will free/release the internal data contained
 * within an ATTRIBUTE_VALUES struct.  All stings that it contains
 * will be freed so make copies of any data you want to keep.
 * On error errorStr will, if suplied, be populated with an error string.
 */
SD_RETURN_TYPE ReleaseAttributeValues(ATTRIBUTE_VALUES* values,
                                      char** errorStr);

/* This function will return the URL from a SERVICE_DESCRIPTION structure.
 * On error errorStr will, if suplied, be populated with an error string.
 * The returned string will be allocated on the heap and it will therefore
 * be the user's responsibility to free() it
 */
char* ServiceDescriptionGetUrl(SERVICE_DESCRIPTION* desc, char** errorStr);

/* This function will return the related services from a
 * SERVICE_DESCRIPTION structure.
 * The results will be returned in the supplied SERVICE_DESCRIPTION_ARRAY
 * On error errorStr will, if suplied, be populated with an error string.
 * This array will need to be released with a call
 * to ReleaseServiceDescriptionArray() when the user has finished with it
 */
SD_RETURN_TYPE
   ServiceDescriptionGetRelatedServices(SERVICE_DESCRIPTION* desc,
                                        SERVICE_DESCRIPTION_ARRAY* arr,
                                        char** errorStr);

/* This function will return the SERVICE_DATA structure from a
 * SERVICE_DESCRITPTION structure.
 * On error errorStr will, if suplied, be populated with an error string.
 * The returned SERVICE_DATA will be allocated om the heap and therefore
 * ReleaseServiceData() followed by free() will have to be called
 * when the user has finished with it
 */
SERVICE_DATA* ServiceDescriptionGetData(SERVICE_DESCRIPTION* desc,
                                        char** errorStr);

/* This function will free/release the internal data contained
 * within a SERVICE_DATA struct.
 * On error errorStr will, if suplied, be populated with an error string.
 */
SD_RETURN_TYPE ReleaseServiceData(SERVICE_DATA* data, char** errorStr);

/* This function will return a copy of the given SERVICE_DATA struct.
 * This is the only acceptable method of copying such a structure.
 * On error errorStr will, if suplied, be populated with an error string.
 * The returned structure will be allocated on the heap and therefore
 * ReleaseServiceData() followed by free() must be called
 * when the user has finished using it.
 */
SERVICE_DATA* CopyServiceData(const SERVICE_DATA* data, char** errorStr);

/* This function returns an ATTRIBUTE_NAMES structure for the
 * requested SERVICE_DATA
 * On error errorStr will, if suplied, be populated with an error string.
 * The results will be returned in the supplied ATTRIBUTE_NAMES
 * This structure will need to be released with a call
 * to ReleaseAttributesArray() when the user has finished with it
 */
SD_RETURN_TYPE ServiceDataListAttributes(SERVICE_DATA* data,
                                         ATTRIBUTE_NAMES* arr,
                                         char** errorStr);

/* This function returns all the values of a given attribute
 * for a given SERVICE_DATA
 * On error errorStr will, if suplied, be populated with an error string.
 * The results will be returned in the supplied ATTRIBUTE_VALUES
 * This structure will need to be released with a call
 * to ReleaseAttributeValues() when the user has finished with it
 */
SD_RETURN_TYPE ServiceDataGetAttribute(SERVICE_DATA* data,
                                       const char* attribute,
                                       ATTRIBUTE_VALUES* values,
                                       char** error_str);

#ifdef __cplusplus
}
#endif

#endif
