/*
 * Copyright (c) Members of the EGEE Collaboration. 2009-2010.
 * See http://www.eu-egee.org/partners/ for details on the copyright
 * holders.
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 *     http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include "sd.h"
#include "saga/saga.hpp"
#include "saga/saga/sd.hpp"
#include <iostream>

using saga::sd::discoverer;
using saga::sd::service_description;
using saga::sd::service_data;

namespace
{
   //Helper function that takes a string and copies it into
   //an allocated char* referenced by a char**
   SD_RETURN_TYPE CreateErrorString(char** errorStr,
                                    const std::string& message)
   {
#ifdef SD_DEBUG
      std::cerr << message << std::endl;
#endif

      if ( errorStr == NULL )
      {
         return SD_OK;
      }

      //We use calloc to ensure zeros
      //Remember to add 1 for the \0
      *errorStr = static_cast<char*>(calloc(message.length() + 1,
                                            sizeof(char)));

      if ( *errorStr == NULL )
      {
         //static_cast failed
#ifdef SD_DEBUG
         std::cerr << "CreateErrorString() calloc/static_cast failed"
                   << std::endl;
#endif
         return SD_FAIL;
      }

      else
      {
         (void)strcpy(*errorStr, message.c_str());
      }
      return SD_OK;
   }

   //Helper function to process saga exceptions
   SD_RETURN_TYPE HandleSagaException(const saga::exception& ex,
                                      char** errorStr)
   {
      const std::string message(ex.get_message());

#ifdef SD_DEBUG
      std::cerr << message << std::endl;
#endif

      return CreateErrorString(errorStr, ex.get_message());
   }

SD_RETURN_TYPE
   PopulateServiceDescriptionArray(const std::vector<service_description>& vec,
                                   SERVICE_DESCRIPTION_ARRAY* arr,
                                   char** errorStr)
{
   //How many entries returned?
   arr->count = vec.size();

   if ( arr->count == 0 )
   {
      arr->descriptionArray = NULL;
   }

   else
   {
      arr->descriptionArray = static_cast<SERVICE_DESCRIPTION*>
         (calloc(arr->count, sizeof(SERVICE_DESCRIPTION)));

      //Check our calloc succeeded
      if ( arr->descriptionArray == NULL )
      {
         const std::string message("PopulateServiceDescriptionArray() "
                                   "calloc/static_cast failed");

         CreateErrorString(errorStr, message);
         return SD_FAIL;
      }

      //OK, now populate our array of SERVICE_DESCRIPTIONS
      std::vector<service_description>::size_type count;
      std::vector<service_description>::size_type entries = vec.size();

      try
      {
         for ( count = 0; count < entries; ++count )
         {
            arr->descriptionArray[count]._serviceDescription =
               static_cast<void*>(new service_description(vec[count]));

            if ( arr->descriptionArray[count]._serviceDescription == NULL )
            {
               const std::string message("PopulateServiceDescriptionArray() "
                                         "static_cast failed");

               CreateErrorString(errorStr, message);
               (void)ReleaseServiceDescriptionArray(arr, NULL);
               return SD_FAIL;
            }
         }
      }

      //If anything throws, tidy up the array and exit
      catch (const saga::exception& ex)
      {
         HandleSagaException(ex, errorStr);
         return SD_FAIL;
      }

      catch (...)
      {
         const std::string message("PopulateServiceDescriptionArray() threw");

         CreateErrorString(errorStr, message);
         (void)ReleaseServiceDescriptionArray(arr, NULL);
         return SD_FAIL;
      }
   }
   return SD_OK;
}

};

extern "C"

{
   SD_RETURN_TYPE InitDiscoverer(DISCOVERER* disco, const char* Url,
                                 char** errorStr)
   {
      //If the user passes in NULL
      //then they get what they deserve
      if ( disco == NULL )
      {
         CreateErrorString(errorStr, "DISCOVERER was NULL");
         return SD_FAIL;
      }

      else
      {
         //We can't throw so fail gracefully
         try
         {
            disco->_disco = new discoverer(std::string(Url));
         }

         catch (saga::exception& ex)
         {
            HandleSagaException(ex, errorStr);
            return SD_FAIL;
         }

         catch (...)
         {
            const std::string message("InitDiscoverer() threw");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }
      }
      //All OK
      return SD_OK;
   }

   SD_RETURN_TYPE ReleaseDiscoverer(DISCOVERER* disco, char** errorStr)
   {
      //If the user passes in NULL
      //then they get what they deserve
      if ( disco == NULL )
      {
         CreateErrorString(errorStr, "DISCOVERER was NULL");
         return SD_FAIL;
      }

      else
      {
         if ( disco->_disco != NULL )
         {
            std::string message;

            //If a destructor throws we hunt down
            //the person responsible and make sure
            //that they're truely sorry.
            try
            {
               discoverer* d = reinterpret_cast<discoverer*>(disco->_disco);

               if ( d == NULL )
               {
                  message += "ReleaseDiscoverer() reinterpret_cast failed.";

               }
               delete d;
               disco->_disco = NULL;
            }

            catch (...)
            {
               message += "ReleaseDiscoverer() threw.";
            }

            if ( message.empty() == false )
            {
               CreateErrorString(errorStr, message);
               return SD_FAIL;
            }
         }
      }
      //All OK
      return SD_OK;
   }

   SD_RETURN_TYPE ListServices2(const DISCOVERER* disco,
                                const char* serviceFilter,
                                const char* dataFilter,
                                SERVICE_DESCRIPTION_ARRAY* arr,
                                char** errorStr)
   {
      //If the user passes in NULL, hasn't called InitDiscoverer()
      //or hasn't passed a SERVICE_DESCRIPTION_ARRAY in
      //then they get what they deserve
      if ( (disco == NULL) ||
           (disco->_disco == NULL) ||
           (arr == NULL) )
      {
         CreateErrorString(errorStr, "NULL value passed");
         return SD_FAIL;
      }

      else
      {
         std::vector<service_description> vec;

         try
         {
            discoverer* d = reinterpret_cast<discoverer*>(disco->_disco);

            if ( d == NULL )
            {
               const std::string message("ListServices2() "
                                         "reinterpret_cast failed");

               CreateErrorString(errorStr, message);
               return SD_FAIL;
            }

            vec = d->list_services(std::string(serviceFilter),
                                   std::string(dataFilter));
         }

         //We can't throw so fail gracefully
         catch (saga::exception& ex)
         {
            HandleSagaException(ex, errorStr);
            return SD_FAIL;
         }

         catch (...)
         {
            CreateErrorString(errorStr, "ListServices2() threw");
            return SD_FAIL;
         }

         return PopulateServiceDescriptionArray(vec, arr, errorStr);
      }

      //All OK
      return SD_OK;
   }

   SD_RETURN_TYPE ListServices3(const DISCOVERER* disco,
                                const char* serviceFilter,
                                const char* authzFilter,
                                const char* dataFilter,
                                SERVICE_DESCRIPTION_ARRAY* arr,
                                char** errorStr)
   {
      //If the user passes in NULL, hasn't called InitDiscoverer()
      //or hasn't passed a SERVICE_DESCRIPTION_ARRAY in
      //then they get what they deserve
      if ( (disco == NULL) ||
           (disco->_disco == NULL) ||
           (arr == NULL) )
      {
         CreateErrorString(errorStr, "NULL value passed");
         return SD_FAIL;
      }

      else
      {
         std::vector<service_description> vec;

         try
         {
            discoverer* d = reinterpret_cast<discoverer*>(disco->_disco);

            if ( d == NULL )
            {
               const std::string message("ListServices3() "
                                         "reinterpret_cast failed");

               CreateErrorString(errorStr, message);
               return SD_FAIL;
            }

            vec = d->list_services(std::string(serviceFilter),
                                   std::string(authzFilter),
                                   std::string(dataFilter));
         }

         //We can't throw so fail gracefully
         catch (saga::exception& ex)
         {
            HandleSagaException(ex, errorStr);
            return SD_FAIL;
         }

         catch (...)
         {
            CreateErrorString(errorStr, "ListServices3() threw");
            return SD_FAIL;
         }

         return PopulateServiceDescriptionArray(vec, arr, errorStr);
      }

      //All OK
      return SD_OK;
   }

   SD_RETURN_TYPE
      ReleaseServiceDescriptionArray(SERVICE_DESCRIPTION_ARRAY* arr,
                                     char** errorStr)
   {
      //Ignore a NULL parameter as this might be being called
      //by a cleanup routine
      if ( arr == NULL )
      {
         CreateErrorString(errorStr, "SERVICE_DESCRIPTION_ARRAY was NULL");
         return SD_OK;
      }

      //Deal with a missing array
      if ( arr->descriptionArray == NULL )
      {
         arr->count = 0;
         return SD_OK;
      }

      std::string message;

      for ( unsigned int count = 0; count < arr->count; ++count )
      {
         void* sdPtr = arr->descriptionArray[count]._serviceDescription;
         try
         {
            service_description* sd =
               reinterpret_cast<service_description*>(sdPtr);

            if ( sd == NULL )
            {
               message += "ReleaseServiceDescriptionArray() "
                          "reinterpret_cast failed.";
            }

            //Now delete the service description
            delete sd;
         }

         catch (...)
         {
            message += "ReleaseServiceDescriptionArray() threw.";
         }

         //NULL the element in the array
         arr->descriptionArray[count]._serviceDescription = NULL;
      }

      //Now free the entire array
      free(arr->descriptionArray);
      arr->descriptionArray = NULL;
      arr->count = 0;

      //Check if we had any errors
      if ( message.empty() == false )
      {
         CreateErrorString(errorStr, message);
         return SD_FAIL;
      }

      else
      {
         return SD_OK;
      }
   }

   SERVICE_DESCRIPTION*
      CopyServiceDescription(const SERVICE_DESCRIPTION* desc,
                             char** errorStr)
   {
      //If the user gives us nothing they get nothing
      if ( desc == NULL )
      {
         CreateErrorString(errorStr, "SERVICE_DESCRIPTION was NULL");
         return NULL;
      }

      //Allocate space for a new SERVICE_DESCRIPTION
      //We use calloc() as it zeros memory for us
      SERVICE_DESCRIPTION* retDesc = static_cast<SERVICE_DESCRIPTION*>
         (calloc(1, sizeof(SERVICE_DESCRIPTION)));

      /*If calloc() failed then return NULL*/
      if ( retDesc == NULL )
      {
         const std::string message("CopyServiceDescription() "
                                   "calloc/static_cast failed");

         CreateErrorString(errorStr, message);
         return NULL;
      }

      //If the provided SERVICE_DESCRIPTION contained a service_description
      //then copy it
      if ( desc->_serviceDescription != NULL )
      {
         try
         {
            //Cast our void* to the correct type
            service_description* sd = reinterpret_cast
               <service_description*>(desc->_serviceDescription);

            if ( sd == NULL )
            {
               const std::string message("CopyServiceDescription() "
                                         "reinterpret_cast failed");

               CreateErrorString(errorStr, message);
               return NULL;
            }

            else
            {
               //Create our copy
               retDesc->_serviceDescription =
                  static_cast<void*>(new service_description(*sd));

               if ( retDesc->_serviceDescription == NULL )
               {
                  const std::string message("CopyServiceDescription() "
                                            "alloc/static_cast failed");

                  CreateErrorString(errorStr, message);
                  return NULL;
               }
            }
         }

         //We can't throw so fail gracefully
         catch (saga::exception& ex)
         {
            HandleSagaException(ex, errorStr);
         }

         catch (...)
         {
            CreateErrorString(errorStr, "CopyServiceDescription() threw");
         }
      }
      return retDesc;
   }

   SD_RETURN_TYPE ReleaseServiceDescription(SERVICE_DESCRIPTION* desc,
                                            char** errorStr)
   {
      SD_RETURN_TYPE retVal = SD_OK;

      //Ignore a NULL parameter as this might be being called
      //by a cleanup routine
      if ( desc == NULL )
      {
         return SD_OK;
      }

      void* sdPtr = desc->_serviceDescription;

      if ( sdPtr != NULL )
      {
         try
         {
            service_description* sd =
               reinterpret_cast<service_description*>(sdPtr);

            if ( sd == NULL )
            {
               const std::string message("ReleaseServiceDescription() "
                                         "reinterpret_cast failed");

               CreateErrorString(errorStr, message);
               retVal = SD_FAIL;
            }

            //Now delete the service description
            delete sd;
         }

         //We can't throw so fail gracefully
         catch (...)
         {
            CreateErrorString(errorStr, "ReleaseServiceDescription() threw");
            retVal = SD_FAIL;
         }
      }

      //NULL the pointer in the SERVICE_DESCRIPTION
      desc->_serviceDescription = NULL;

      return retVal;
   }

   SD_RETURN_TYPE ServiceDescriptionListAttributes(SERVICE_DESCRIPTION* desc,
                                                   ATTRIBUTE_NAMES* arr,
                                                   char** errorStr)
   {
      //If the user gives us nothing they get nothing
      if ( (desc == NULL) ||
           (desc->_serviceDescription == NULL) ||
           (arr == NULL) )
      {
         CreateErrorString(errorStr, "Parameter was NULL");
         return SD_FAIL;
      }

      try
      {
         service_description* sd =
            reinterpret_cast<service_description*>(desc->_serviceDescription);

         if ( sd == NULL )
         {
            const std::string message("ServiceDescriptionListAttributes() "
                                      "reinterpret_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         std::vector<std::string> atts = sd->list_attributes();

         //Copy the attribute count to our structure
         std::vector<std::string>::size_type count;
         std::vector<std::string>::size_type entries = atts.size();
         arr->count = entries;

         //Now allocate an array big enough
         //to contain all the attributes
         arr->attributeNames =
            static_cast<char**>(calloc(arr->count, sizeof(char*)));

         if ( arr->attributeNames == NULL )
         {
            const std::string message("ServiceDescriptionListAttributes() "
                                      "calloc/static_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         //Loop through each attribute copying it to our structure
         for ( count = 0; count < entries; ++count )
         {
            const char* att = atts[count].c_str();

            //We use calloc to ensure zeros
            //Remember to add 1 for the \0
            arr->attributeNames[count] =
               static_cast<char*>(calloc(strlen(att) + 1, sizeof(char)));

            if ( arr->attributeNames[count] != NULL )
            {
               (void)strcpy(arr->attributeNames[count], att);
            }
         }
      }

      //If anything throws, tidy up the array and exit
      catch (saga::exception& ex)
      {
         HandleSagaException(ex, errorStr);
         (void)ReleaseAttributeNames(arr, NULL);
         return SD_FAIL;
      }

      catch (...)
      {
         CreateErrorString(errorStr, 
                           "ServiceDescriptionListAttributes() threw");

         (void)ReleaseAttributeNames(arr, NULL);
         return SD_FAIL;
      }
      return SD_OK;
   }

   SD_RETURN_TYPE ReleaseAttributeNames(ATTRIBUTE_NAMES* arr,
                                        char** errorStr)
   {
      //Ignore a NULL parameter as this might be being called
      //by a cleanup routine
      if ( arr == NULL )
      {
         return SD_OK;
      }

      //Deal with a missing array
      if ( arr->attributeNames == NULL )
      {
         arr->count = 0;
         return SD_OK;
      }

      std::string message;

      //Loop through every string in the aray freeing them
      for ( unsigned int count = 0; count < arr->count; ++count )
      {
         try
         {
            char* att  = static_cast<char*>(arr->attributeNames[count]);

            if ( att == NULL )
            {
               message += "ReleaseAttributeNames() static_cast failed.";
            }

            //Now delete the attribute
            free(att);
         }

         //We can't throw so fail gracefully
         catch (...)
         {
            message += "ReleaseAttributeNames() threw.";
         }

         //NULL the element in the array
         arr->attributeNames[count] = NULL;
      }

      //Now free the entire array
      free(arr->attributeNames);
      arr->attributeNames = NULL;
      arr->count = 0;

      if ( message.empty() == false )
      {
         CreateErrorString(errorStr, message);
         return SD_FAIL;
      }

      else
      {
         return SD_OK;
      }
   }

   SD_RETURN_TYPE ServiceDescriptionGetAttribute(SERVICE_DESCRIPTION* desc,
                                                 const char* attribute,
                                                 ATTRIBUTE_VALUES* values,
                                                 char** errorStr)
   {
      //If the user give us nothing they get nothing
      if ( (desc == NULL) ||
           (desc->_serviceDescription == NULL) ||
           (attribute == NULL) ||
           (values == NULL) )
      {
         CreateErrorString(errorStr, "Parameter was NULL");
         return SD_FAIL;
      }

      try
      {
         service_description* sd =
            reinterpret_cast<service_description*>(desc->_serviceDescription);

         if ( sd == NULL )
         {
            const std::string message("ServiceDescriptionGetAttribute() "
                                      "reinterpret_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         std::vector<std::string> valuesVec;

         //If the attribute is a vector we need to store many values
         if ( sd->attribute_is_vector(std::string(attribute)) )
         {
            //Get the values
            valuesVec = sd->get_vector_attribute(attribute);
         }

         //Otherwise there's only one value
         else
         {
            std::string value = sd->get_attribute(attribute);
            valuesVec.push_back(value);
         }

         //Copy the attribute count to our structure
         std::vector<std::string>::size_type count;
         std::vector<std::string>::size_type entries = valuesVec.size();
         values->count = entries;

         //Now allocate an array big enough
         //to contain all the values
         values->attributeValues =
            static_cast<char**>(calloc(values->count, sizeof(char*)));

         if ( values->attributeValues == NULL )
         {
            const std::string message("ServiceDescriptionGetAttribute() "
                                      "calloc/static_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         //Loop through each attribute copying it to our structure
         for ( count = 0; count < entries; ++count )
         {
            const char* val = valuesVec[count].c_str();

            //We use calloc to ensure zeros
            //Remember to add 1 for the \0
            values->attributeValues[count] =
            static_cast<char*>(calloc(strlen(val) + 1, sizeof(char)));

            if ( values->attributeValues[count] != NULL )
            {
               (void)strcpy(values->attributeValues[count], val);
            }
         }
      }

      catch (...)
      {
         CreateErrorString(errorStr,
                           "ServiceDescriptionGetAttribute() threw()");
         (void)ReleaseAttributeValues(values, NULL);
         return SD_FAIL;
      }
      return SD_OK;
   }

   SD_RETURN_TYPE ReleaseAttributeValues(ATTRIBUTE_VALUES* values,
                                         char** errorStr)
   {
      //Ignore a NULL parameter as this might be being called
      //by a cleanup routine
      if ( values == NULL )
      {
         return SD_OK;
      }

      //Deal with a missing array
      if ( values->attributeValues == NULL )
      {
         values->count = 0;
         return SD_OK;
      }

      std::string message;

      //Loop through every string in the aray freeing them
      for ( unsigned int count = 0; count < values->count; ++count )
      {
         try
         {
            char* value = static_cast<char*>(values->attributeValues[count]);

            if ( values == NULL )
            {
               message += "ReleaseAttributeValues() static_cast failed.";
            }

            //Now delete the value
            free(value);
         }

         //We can't throw so fail gracefully
         catch (...)
         {
            message += "ReleaseAttributeValues() threw.";
         }

         //NULL the element in the array
         values->attributeValues[count] = NULL;
      }

      //Now free the entire array
      free(values->attributeValues);
      values->attributeValues = NULL;
      values->count = 0;

      if ( message.empty() == false )
      {
         CreateErrorString(errorStr, message);
         return SD_FAIL;
      }

      else
      {
         return SD_OK;
      }
   }

   char* ServiceDescriptionGetUrl(SERVICE_DESCRIPTION* desc,
                                  char** errorStr)
   {
      char* retStr = NULL;

      //If the user gives us nothing they get nothing
      if ( (desc == NULL) ||
           (desc->_serviceDescription == NULL) )
      {
         CreateErrorString(errorStr, "SERVICE_DESCRIPTION was NULL/empty");
         return NULL;
      }

      try
      {
         //Get the service_description out of our struct
         service_description* sd =
            reinterpret_cast<service_description*>(desc->_serviceDescription);

         if ( sd == NULL )
         {
            const std::string message("ServiceDescriptionGetUrl() "
                                      "reinterpret_cast failed");

            CreateErrorString(errorStr, message);
            return NULL;
         }

         //Copy the URL into our own char*
         std::string urlStr = sd->get_url();
         const char* url = urlStr.c_str();

         //We use calloc to ensure zeros
         //Remember to add 1 for the \0
         retStr = static_cast<char*>(calloc(strlen(url) + 1, sizeof(char)));
         (void)strcpy(retStr, url);
      }

      //We can't throw so fail gracefully
      //Fall trhough OK as retStr will be in a sensible state
      catch (saga::exception& ex)
      {
         HandleSagaException(ex, errorStr);
      }

      catch (...)
      {
         CreateErrorString(errorStr, "ServiceDescriptionGetUrl() threw");
      }

      return retStr;
   }

   SD_RETURN_TYPE
      ServiceDescriptionGetRelatedServices(SERVICE_DESCRIPTION* desc,
                                           SERVICE_DESCRIPTION_ARRAY* arr,
                                           char** errorStr)
   {
      //If the user gives us nothing they get nothing
      if ( (desc == NULL) ||
           (desc->_serviceDescription == NULL) ||
           (arr == NULL) )
      {
         CreateErrorString(errorStr, "Parameter was NULL/empty");
         return SD_FAIL;
      }

      //Get arr into a sensible state
      arr->descriptionArray = NULL;

      try
      {
         //Get the service_description out of our struct
         service_description* sd =
            reinterpret_cast<service_description*>(desc->_serviceDescription);

         if ( sd == NULL )
         {
            const std::string
               message("ServiceDescriptionGetRelatedServices() "
                       "reinterpret_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         //Get the related services if they exist
         std::vector<service_description> relatedVec;

         if ( sd->attribute_exists(
                 saga::sd::attributes::service_description_relatedservices) )
         {
            relatedVec = sd->get_related_services();
         }

         //Now populate our array with the related services
         //doesn't matter if our vector is empty
         PopulateServiceDescriptionArray(relatedVec, arr, errorStr);
      }

      //We can't throw so fail gracefully
      catch (saga::exception& ex)
      {
         HandleSagaException(ex, errorStr);
         ReleaseServiceDescriptionArray(arr, NULL);
         return SD_FAIL;
      }
      catch (...)
      {
         CreateErrorString(errorStr,
                           "ServiceDescriptionGetRelatedServices() threw");

         ReleaseServiceDescriptionArray(arr, NULL);
         return SD_FAIL;
      }
      return SD_OK;
   }

   SERVICE_DATA* ServiceDescriptionGetData(SERVICE_DESCRIPTION* desc,
                                           char** errorStr)
   {
      //If the user gives us nothing they get nothing
      if ( (desc == NULL) ||
           (desc->_serviceDescription == NULL) )
      {
         CreateErrorString(errorStr, "Parameter was NULL/empty");
         return NULL;
      }

      SERVICE_DATA* retData = NULL;
      service_data* servData = NULL;

      try
      {
         //Get the service_description out of our struct
         service_description* sd =
            reinterpret_cast<service_description*>(desc->_serviceDescription);

         if ( sd == NULL )
         {
            const std::string message("ServiceDescriptionGetData() "
                                      "reinterpret_cast failed");

            CreateErrorString(errorStr, message);
            return NULL;
         }

         //Allocate space for a new SERVICE_DATA
         //We use calloc() as it zeros memory for us
         retData =
            static_cast<SERVICE_DATA*>(calloc(1, sizeof(SERVICE_DATA)));

         /*If calloc() failed then return NULL*/
         if ( retData == NULL )
         {
            const std::string message("ServiceDescriptionGetData() "
                                      "calloc/static_cast failed");

            CreateErrorString(errorStr, message);
            return NULL;
         }

         //Create a new service_data object
         servData = new service_data();
         if ( servData != NULL )
         {
            *servData = sd->get_data();
         }
         
         //Populate our structure
         retData->_serviceData = static_cast<void*>(servData);
      }

      //We can't throw so fail gracefully
      catch (...)
      {
         //Free any allocated memory
         if ( retData != NULL )
         {
            free(retData);
         }

         if ( servData != NULL )
         {
            delete servData;
         }

         CreateErrorString(errorStr,
                           "ServiceDescriptionGetData() threw");
         return NULL;
      }

      return retData;
   }

   SD_RETURN_TYPE ReleaseServiceData(SERVICE_DATA* data, char** errorStr)
   {
      //Ignore a NULL parameter as this might be being called
      //by a cleanup routine
      if ( (data == NULL) ||
           (data->_serviceData == NULL) )
      {
         return SD_OK;
      }

      void* servDataPtr = data->_serviceData;

      std::string message;

      try
      {
         service_data* servData =
            reinterpret_cast<service_data*>(servDataPtr);

         if ( servData == NULL )
         {
            message += "ReleaseServiceData() reinterpret_cast failed.";
         }

         //Now delete the service data
         delete servData;
      }

      //We can't throw so fail gracefully
      catch (...)
      {
         message += "ReleaseServiceData() threw.";
      }

      //NULL the pointer in the SERVICE_DESCRIPTION
      data->_serviceData = NULL;

      if ( message.empty() == false )
      {
         CreateErrorString(errorStr, message);
         return SD_FAIL;
      }

      else
      {
         return SD_OK;
      }
   }

   SERVICE_DATA* CopyServiceData(const SERVICE_DATA* data, char** errorStr)
   {
      //If the user gives us nothing they get nothing
      if ( data == NULL )
      {
         CreateErrorString(errorStr, "SERVICE_DATA was NULL");
         return NULL;
      }

      //Allocate space for a new SERVICE_DATA
      //We use calloc() as it zeros memory for us
      SERVICE_DATA* retData = static_cast<SERVICE_DATA*>
         (calloc(1, sizeof(SERVICE_DATA)));

      /*If calloc() failed then return NULL*/
      if ( retData == NULL )
      {
         const std::string message("CopyServiceData() "
                                   "calloc/static_cast failed");

         CreateErrorString(errorStr, message);
         return NULL;
      }

      //If the provided SERVICE_DATA contained a service_data
      //then copy it
      if ( data->_serviceData != NULL )
      {
         try
         {
            //Cast our void* to the correct type
            service_data* servData = reinterpret_cast
               <service_data*>(data->_serviceData);

            if ( servData == NULL )
            {
               const std::string message("CopyServiceData() "
                                         "reinterpret_cast failed");

               CreateErrorString(errorStr, message);
               free(retData);
               return NULL;
            }

            else
            {
               //Create our copy
               retData->_serviceData =
                  static_cast<void*>(new service_data(*servData));

               if ( retData->_serviceData == NULL )
               {
                  const std::string message("CopyServiceData() "
                                            "calloc/static_cast failed");

                  CreateErrorString(errorStr, message);
                  free(retData);
                  return NULL;
               }
            }
         }

         //We can't throw so fail gracefully
         catch (...)
         {
            CreateErrorString(errorStr, "CopyServiceData() threw");
            free(retData);
            return NULL;
         }
      }
      return retData;
   }

   SD_RETURN_TYPE ServiceDataListAttributes(SERVICE_DATA* data,
                                            ATTRIBUTE_NAMES* arr,
                                            char** errorStr)
   {
      //If the user gives us nothing they get nothing
      if ( (data == NULL) ||
           (data->_serviceData == NULL) ||
           (arr == NULL) )
      {
         CreateErrorString(errorStr, "Parameter was NULL/empty");
         return SD_FAIL;
      }

      try
      {
         service_data* servData =
            reinterpret_cast<service_data*>(data->_serviceData);

         if ( servData == NULL )
         {
            const std::string message("ServiceDataListAttributes() "
                                      "reinterpret_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         std::vector<std::string> atts = servData->list_attributes();

         //Copy the attribute count to our structure
         std::vector<std::string>::size_type count;
         std::vector<std::string>::size_type entries = atts.size();
         arr->count = entries;

         //Now allocate an array big enough
         //to contain all the attributes
         arr->attributeNames =
            static_cast<char**>(calloc(arr->count, sizeof(char*)));

         if ( arr->attributeNames == NULL )
         {
            const std::string message("ServiceDataListAttributes() "
                                      "calloc/static_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         //Loop through each attribute copying it to our structure
         for ( count = 0; count < entries; ++count )
         {
            const char* att = atts[count].c_str();

            //We use calloc to ensure zeros
            //Remember to add 1 for the \0
            arr->attributeNames[count] =
               static_cast<char*>(calloc(strlen(att) + 1, sizeof(char)));

            if ( arr->attributeNames[count] != NULL )
            {
               (void)strcpy(arr->attributeNames[count], att);
            }
         }
      }

      //If anything throws, tidy up the array and exit
      catch (saga::exception& ex)
      {
         HandleSagaException(ex, errorStr);
         (void)ReleaseAttributeNames(arr, NULL);
         return SD_FAIL;
      }

      catch (...)
      {
         CreateErrorString(errorStr, "ServiceDataListAttributes() threw");
         (void)ReleaseAttributeNames(arr, NULL);
         return SD_FAIL;
      }
      return SD_OK;
   }

   SD_RETURN_TYPE ServiceDataGetAttribute(SERVICE_DATA* data,
                                                 const char* attribute,
                                                 ATTRIBUTE_VALUES* values,
                                                 char** errorStr)
   {
      //If the user give us nothing they get nothing
      if ( (data == NULL) ||
           (data->_serviceData == NULL) ||
           (attribute == NULL) ||
           (values == NULL) )
      {
         CreateErrorString(errorStr, "Parameter was NULL/empty");
         return SD_FAIL;
      }

      try
      {
         service_data* servData =
            reinterpret_cast<service_data*>(data->_serviceData);

         if ( servData == NULL )
         {
            const std::string message("ServiceDataGetAttribute() "
                                      "reinterpret_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         std::vector<std::string> valuesVec;

         //If the attribute is a vector we need to store many values
         if ( servData->attribute_is_vector(std::string(attribute)) )
         {
            //Get the values
            valuesVec = servData->get_vector_attribute(attribute);
         }

         //Otherwise there's only one value
         else
         {
            std::string valueStr = servData->get_attribute(attribute);
            valuesVec.push_back(valueStr);
         }

         //Copy the attribute count to our structure
         std::vector<std::string>::size_type count;
         std::vector<std::string>::size_type entries = valuesVec.size();
         values->count = entries;

         //Now allocate an array big enough
         //to contain all the values
         values->attributeValues =
            static_cast<char**>(calloc(values->count, sizeof(char*)));

         if ( values->attributeValues == NULL )
         {
            const std::string message("ServiceDataGetAttribute() "
                                      "calloc/static_cast failed");

            CreateErrorString(errorStr, message);
            return SD_FAIL;
         }

         //Loop through each attribute copying it to our structure
         for ( count = 0; count < entries; ++count )
         {
            const char* val = valuesVec[count].c_str();

            //We use calloc to ensure zeros
            //Remember to add 1 for the \0
            values->attributeValues[count] =
            static_cast<char*>(calloc(strlen(val) + 1, sizeof(char)));

            if ( values->attributeValues[count] != NULL )
            {
               (void)strcpy(values->attributeValues[count], val);
            }
         }
      }

      //We can't throw so fail gracefully
      catch (saga::exception& ex)
      {
         HandleSagaException(ex, errorStr);
         (void)ReleaseAttributeValues(values, NULL);
         return SD_FAIL;
      }

      catch (...)
      {
         CreateErrorString(errorStr, "ServiceDataGetAttribute() threw()");
         (void)ReleaseAttributeValues(values, NULL);
         return SD_FAIL;
      }
      return SD_OK;
   }

} // extern "C"

